import es
import langlib
import popuplib
import playerlib
import gamethread
import vecmath
import effectlib
import services
import time
from vecmath import vector
from configobj import ConfigObj
import psyco
psyco.full()

info = es.AddonInfo()
info.name = 'SpawnConquer'
info.basename = 'scq'
info.version = '1.0'
info.author = 'GODJonez'
info.description = 'Capture spawn areas to conquer the map and win'
info.tags = 'gameplay mod spawn deathmatch'

# textdata will hold the information in multiple languages using langlib
textdata = None

rules = {}
rules['winning'] = es.ServerVar('scq_rules_winning', 0,
    '0 = capture main spawn area, 1 = kill opposing team, 2 = capture all spawns')
rules['spawns'] = es.ServerVar('scq_rules_spawns', 2,
    '0 = only map spawns, 1 = capture any spawn, 2 = capture linked, 3 = team linked')

settings = {}
settings['weapons'] = es.ServerVar('scq_clear_weapons', 0,
    'Set to 1 to remove dropped weapons automatically')
settings['ragdolls'] = es.ServerVar('scq_clear_ragdolls', 1,
    'Set to 1 to remove ragdolls automatically')
settings['props'] = es.ServerVar('scq_clear_props', 1,
    'Set to 1 to remove map props')
settings['spawntime'] = es.ServerVar('scq_spawntime', 5.0,
    'Time in seconds until player respawns')
settings['updatetime'] = 2.0
settings['circlesteps'] = 12

sounds = {}
sounds['capture'] = 'hl1/fvox/bell.wav'
sounds['steal'] = 'hl1/fvox/bell.wav'
sounds['victory'] = 'scq/fanfare.mp3'
sounds['defeat'] = 'scq/gameover.mp3'

currentmap = es.server_var['eventscripts_currentmap']
mapdata = None
players = {}
popups = {}
TeamSecurity = {
    2: es.ServerVar('_scq_teamsec2', 0, 'fakeclient userid'),
    3: es.ServerVar('_scq_teamsec3', 0, 'fakeclient userid')
    }
istracking = False

bullet_impact_shield = False

inAdminMode = False
adminInverseLinks = True
adminChosenType = 'box'
adminAdjustStep = 10
adminLinkFrom = None
adminLinkTeam = 0
adminEditingSpawn = None
adminEditingCoord = None
adminCoordBackup = None
adminEditingText = None
adminEditingValue = None
adminAddingSpawn = False

teamcolors = {
    0: (255, 255, 255),
    1: (64, 255, 0),
    2: (255, 0, 0),
    3: (0, 64, 255)
    }

macros = {
    'playerpopup': 'scq_user%d',
    'spawnpopup': 'scq_spawn%d',
    'otherpopup': 'scq_%s',
    'adminpopup': 'scq_admin_%s',
    }

class VersionMismatch(RuntimeError):
    '''
File version not supported
    '''
    def __init__(self, message=None):
        super(self, VersionMismatch).__init__(message)

class Mapdata(object):
    '''
Class containing spawn area data for SpawnConquer
Python script addon for EventScripts
    '''
    def __init__(self, meta, def2, def3):
        '''
Parameter meta must be a dict-like object containing information:
    author
    version
    s_version
        '''
        self.author = meta['author']
        self.version = meta['version']
        self.s_version = meta['s_version']
        self.def2 = def2
        self.def3 = def3
        self.towns = 0
        self.ctowns = 0
        self.data = {}
        if float(self.s_version) > float(info.version):
            raise VersionMismatch, "This config is for newer version of %s"%info.name
    def __setitem__(self, key, value):
        '''
__setitem__(self, key, value) <=> self[key] = value
        '''
        self.data[key] = value
    def __getitem__(self, key):
        '''
x = __getitem__(self, key) <=> x = self[key]
        '''
        return self.data[key]
    def clear(self):
        '''
Clear the map information, retains metadata
        '''
        self.data.clear()

class SpawnArea(object):
    '''
This class represents one of the spawn areas for SpawnConquer script addon
    '''
    def __init__(self, name, data):
        stuff = data.split(";")
        self.name = name
        self.type = stuff[0]
        self.team = 0
        self.default = 0
        self.tlinks = set()
        self.ctlinks = set()
        if self.type == 'box':
            self.coord1 = vector(stuff[1])
            self.coord2 = vector(stuff[2])
            self.updateOrigin()
            max_xn = int((self.coord1[0]-self.coord2[0])/40)
            max_yn = int((self.coord1[1]-self.coord2[1])/40)
            if max_xn == 0 or max_yn == 0:
                es.dbgmsg(0,"SCQ: Spawn %s has invalid size!"%self.name)
            self.xn = 0
            self.yn = 0
            if max_xn < 0:
                self.xn_range = range(max_xn, 0)
            else:
                self.xn_range = range(1, max_xn)
            if max_yn < 0:
                self.yn_range = range(max_yn, 0)
            else:
                self.yn_range = range(1, max_yn)
        elif self.type == 'circle':
            self.origin = vector(stuff[1])
            self.radius = float(stuff[2])
            self.height = float(stuff[3])
            self.an = 0
            self.xn = 0
            max_xn = int(self.radius/40)
            self.xn_range = range(max_xn, 0, -1)
            self.an_range = range(4)
    def updateOrigin(self):
        '''
Updates origin value if box has been moved
        '''
        if self.type == 'box':
            self.origin = (self.coord1+self.coord2)/2
            self.origin[2] = min((self.coord1[2],self.coord2[2]))
    def strData(self):
        '''
Return the data as it was parsed
        '''
        if self.type == 'box':
            return "box;%s;%s"%(
                str(self.coord1),
                str(self.coord2)
                )
        elif self.type == 'circle':
            return "circle;%s;%f;%f"%(
                str(self.origin),
                self.radius,
                self.height
                )
    def strtLinks(self):
        '''
Return tlinks as string
        '''
        r = ""
        for link in self.tlinks:
            r += str(link)+";"
        return r[:-1]
    def strctLinks(self):
        '''
Return ctlinks as string
        '''
        r = ""
        for link in self.ctlinks:
            r += str(link)+";"
        return r[:-1]
    def isin(self, vec):
        '''
Check if a player (passed as a vector) is inside the spawn area
        '''
        if self.type == 'box':
            return vecmath.isbetweenRect(vec, self.coord1, self.coord2)
        elif self.type == 'circle':
            t = vector(vec)
            t[2] = self.origin[2]
            if vecmath.distance(t, self.origin) <= self.radius:
                if vec[2]>=self.origin[2] and vec[2]-self.origin[2] <= self.height:
                    return True
            return False
    def draw(self):
        '''
Create the effect for the spawn area
        '''
        if self.type == 'box':
            effectlib.drawBox(self.coord1, self.coord2,
                seconds = settings['updatetime']+.1,
                red = teamcolors[self.team][0],
                green = teamcolors[self.team][1],
                blue = teamcolors[self.team][2]
                              )
        elif self.type == 'circle':
            effectlib.drawCircle(self.origin, self.radius,
                steps = settings['circlesteps'],
                seconds = settings['updatetime']+.1,
                red = teamcolors[self.team][0],
                green = teamcolors[self.team][1],
                blue = teamcolors[self.team][2]
                                 )
    def drawlinks(self):
        '''
Draw all link lines with their appropriate colours
        '''
        links = list()
        if self.team == 2:
            links = self.tlinks
        elif self.team == 3:
            links = self.ctlinks
        for sp in links:
            effectlib.drawLine(
                self.origin, mapdata[sp].origin,
                seconds = settings['updatetime'],
                red = teamcolors[self.team][0],
                green = teamcolors[self.team][1],
                blue = teamcolors[self.team][2]
                              )
    def capture(self, team):
        '''
Capture this spawn area for another team
        '''
        if CanBeCaptured(self, team):
            if team == 2:
                mapdata.towns += 1
            else:
                mapdata.ctowns += 1
            if self.team == 0:
                # Captured
                for pl in players:
                    text = textdata('%dcapture'%team, {'name': self.name}, players[pl].get('lang'))
                    es.centertell(int(pl), text)
                    es.playsound(int(pl), sounds['capture'], .7)
            else:
                # Stolen
                if team == 2:
                    mapdata.ctowns -= 1
                    nspawn = mapdata.def3
                else:
                    mapdata.towns -= 1
                    nspawn = mapdata.def2
                for pl in players:
                    player = players[pl]
                    if player.defspawn == self.name and player.attributes['teamid'] == self.team:
                        player.defspawn = nspawn
                    text = textdata('%dstole'%team, {'name': self.name}, player.get('lang'))
                    es.centertell(int(pl), text)
                    es.playsound(int(pl), sounds['steal'], .7)
            self.team = team
            if rules['winning'] == 0:
                if team == 2:
                    if mapdata[mapdata.def3] is self:
                        TriggerWin(team)
                elif team == 3:
                    if mapdata[mapdata.def2] is self:
                        TriggerWin(team)
            elif rules['winning'] == 2:
                if mapdata.towns == len(mapdata.data):
                    # T Wins
                    TriggerWin(2)
                elif mapdata.ctowns == len(mapdata.data):
                    # CT Wins
                    TriggerWin(3)
            UpdatePlayerPopups()
            
    def getSpawnLocation(self):
        '''
Return a random spawn location inside this spawn area
        '''
        if self.type == 'box':
            x = self.coord2[0] + 40*self.xn_range[self.xn]
            y = self.coord2[1] + 40*self.yn_range[self.yn]
            self.xn += 1
            if self.xn == len(self.xn_range):
                self.xn = 0
                self.yn += 1
                if self.yn == len(self.yn_range):
                    self.yn = 0
            return vector((x,y,self.origin[2]))
        elif self.type == 'circle':
            multi = (self.an & 2) - 1
            if self.an & 1:
                x = self.origin[0]
                y = self.origin[1] + multi*40*self.xn_range[self.xn]
            else:
                x = self.origin[0] + multi*40*self.xn_range[self.xn]
                y = self.origin[1]
            self.an += 1
            if self.an == len(self.an_range):
                self.an = 0
                self.xn += 1
                if self.xn == len(self.xn_range):
                    self.xn = 0
            return vector((x,y,self.origin[2]))
    def clearLinks(self):
        '''
Clear all links from self
        '''
        self.tlinks.clear()
        self.ctlinks.clear()
    def clearToLinks(self):
        '''
Clear all links pointing to self
        '''
        for sp in mapdata.data:
            for linklist in (mapdata[sp].tlinks, mapdata[sp].ctlinks):
                for ll in set(linklist):
                    if ll == self.name:
                        linklist.remove(ll)
    def rename(self, newname):
        '''
Rename this spawn area and update all links to the new name
        '''
        # Check if this name already exists!
        for sp in mapdata.data:
            if sp == newname:
                return False
        # Update all links
        for sp in mapdata.data:
            if sp != self.name:
                for linklist in (mapdata[sp].tlinks, mapdata[sp].ctlinks):
                    for ll in set(linklist):
                        if ll == self.name:
                            linklist.remove(ll)
                            linklist.add(newname)
        # update mapdata
        if mapdata.def2 == self.name:
            mapdata.def2 = newname
        if mapdata.def3 == self.name:
            mapdata.def3 = newname
        del mapdata.data[self.name]
        mapdata[newname] = self
        self.name = newname
        return True
                
        
def load():
    '''
This function is automatically run when this script is loaded using es_load
    '''
    global textdata, mapdata, players
    if not es.exists('variable', 'est_version'):
        es.dbgmsg(0,"%s requires ES_Tools to work correctly!"%info.name)
        es.dbgmsg(0,"Install ES_Tools to the server.")
        es.soon('es_unload %s'%info.basename)
    else:
        
        if not services.isRegistered('auth'):
            es.dbgmsg(0,"%s requires auth provider to work correctly!"%info.name)
        else:
            auth = services.use('auth')
            auth.registerCapability("edit_spawnconquer", auth.ADMIN)
            es.addons.registerSayFilter(chat_monitor)
        
        textdata = langlib.Strings(es.getAddonPath(info.basename) + "/textdata.ini")

        if str(currentmap):
            # A map has been started already:
            mapdata = LoadMapdata(str(currentmap))
            for pl in playerlib.getPlayerList():
                players[int(pl)] = pl
            ResetGame()
            CreatePlayerPopups()
            CreateRulesPopup(rules)
            StartTracking()

        es.msg("%s v%s has been loaded."%(info.name, info.version))

def unload():
    '''
This function is automatically run when this script is unloaded using es_unload
    '''
    global textdata, mapdata, players
    if es.exists('variable', 'est_version'):
        DeleteRulesPopup()
        DeletePlayerPopups()
        StopTracking()
        DeleteFakeClients()
        mapdata = None
        players = None
        textdata = None

        for p in popups:
            p.delete()
        
        es.addons.unregisterSayFilter(chat_monitor)
        es.msg("%s v%s has been unloaded."%(info.name, info.version))
    

def es_map_start(event_var):
    '''
This function is called when the server loads a new map
    '''
    global inAdminMode, mapdata
    StopTracking()
    mapdata = LoadMapdata(str(currentmap))
    players.clear()
    AddDownloadables([("sound/%s"%x) for x in sounds.values()])
    TeamSecurity[2].set(0)
    inAdminMode = False

def round_start(event_var):
    '''
This function is called when a new round starts
    '''
    ResetGame()
    DeleteMapObjectives()
    DeleteProps()
    CheckFakeClients()
    CreatePlayerPopups()
    CreateRulesPopup(rules)
    StartTracking()

def round_end(event_var):
    '''
This function is called when a round ends
    '''
    DeletePlayerPopups()
    StopTracking()


def player_team(event_var):
    '''
This function is called whenever a player changes team or disconnects
    '''
    if mapdata:
        userid = int(event_var['userid'])
        team = int(event_var['team'])
        disconnect = int(event_var['disconnect'])
        if disconnect:
            if players.has_key(userid):
                DeletePlayerPopup(userid)
                del players[userid]
        else:
            if not players.has_key(userid):
                players[userid] = playerlib.getPlayer(userid)
            players[userid].attributes['teamid'] = team
            if team > 1:
                if team == 2:
                    players[userid].defspawn = mapdata.def2
                else:
                    players[userid].defspawn = mapdata.def3
                CreatePlayerPopup(userid)
                if int(event_var['oldteam']) < 2:
                    gamethread.delayed(settings['spawntime'], SpawnPlayer, userid)
            elif team == 1:
                DeletePlayerPopup(userid)
                popups[macros['otherpopup']%'spec'].send(userid)

def player_spawn(event_var):
    '''
This function is called whenever the player spawns (as Source engine defines)
    '''
    if mapdata:
        userid = int(event_var['userid'])
        if userid in players:
            user = players[userid]
            if user.attributes['teamid'] > 1:
                pname = macros['playerpopup']%userid
                if not popuplib.isqueued(pname,userid):
                    gamethread.delayed(1, gimmepopup, (pname, userid))
                else:
                    popups[pname].update(userid)
                es.setplayerprop(userid, "CBaseEntity.m_CollisionGroup", 2)

def gimmepopup(pname, userid):
    '''
A wrapper function for popuplib.send to suppress errors
    '''
    try:
        popuplib.send(pname, userid)
    except ValueError:
        # popup doesn't exist = between round_end and round_start
        pass
    except playerlib.UseridError:
        # userid doesn't exist anymore
        if players.has_key(userid):
            del players[userid]

def player_death(event_var):
    '''
This function is called whenever a player dies
    '''
    gamethread.delayed(settings['spawntime'], SpawnPlayer, int(event_var['userid']))
    DeleteRagdolls(int(event_var['userid']))
    DeleteWeapons()


def LoadMapdata(mapname):
    '''
Load data for the map
    '''
    data = ConfigObj("%s/mapdata/%s.ini"%(es.getAddonPath(info.basename),mapname))
    try:
        meta = data['meta']
        spawns = data['spawns']
        tlinks = data['tlinks']
        ctlinks = data['ctlinks']
        defs = data['defaults']
        mp = Mapdata(meta, defs['2'], defs['3'])
    except KeyError:
        es.dbgmsg(0, 'SpawnConquer Data for map "%s" not found'%mapname)
        return None
    except VersionMismatch, e:
        es.dbgmsg(0, 'SpawnConquer Data for map "%s" is wrong version'%mapname)
        es.dbgmsg(0, e)
        return None

    for spawnarea in spawns:
        mp[spawnarea] = SpawnArea(spawnarea, spawns[spawnarea])

    for link in tlinks:
        if str(tlinks[link]):
            mp[link].tlinks = tlinks[link].split(';')
    for link in ctlinks:
        if str(ctlinks[link]):
            mp[link].ctlinks = ctlinks[link].split(';')

    return mp


def SaveMapdata(data, mapname):
    '''
Save data of the map
    '''
    datafile = open("%s/mapdata/%s.ini"%(es.getAddonPath(info.basename),mapname), "w")
    datafile.write('[meta]\n')
    datafile.write('author="%s"\n'%data.author)
    datafile.write('version="%s"\n'%data.version)
    datafile.write('s_version="%s"\n'%info.version)
    datafile.write('\n')
    datafile.write('[spawns]\n')
    for sp in data.data:
        spawn = data[sp]
        datafile.write('%s="%s"\n'%(sp, spawn.strData()))
    datafile.write('\n')
    datafile.write('[tlinks]\n')
    for sp in data.data:
        spawn = data[sp]
        datafile.write('%s="%s"\n'%(sp, spawn.strtLinks()))
    datafile.write('\n')
    datafile.write('[ctlinks]\n')
    for sp in data.data:
        spawn = data[sp]
        datafile.write('%s="%s"\n'%(sp, spawn.strctLinks()))
    datafile.write('\n')
    datafile.write('[defaults]\n')
    datafile.write('2="%s"\n'%data.def2)
    datafile.write('3="%s"\n'%data.def3)
    datafile.flush()
    datafile.close()
    return True


def ResetGame():
    '''
Reset SpawnConquer
    '''
    # Reset all spawn areas
    if mapdata:
        for sp in mapdata.data:
            mapdata[sp].team = 0
        
        mapdata[mapdata.def2].team = 2
        mapdata[mapdata.def3].team = 3

        mapdata.towns = 1
        mapdata.ctowns = 1
        
        # Reset all players
        for pl in players:
            if players[pl].attributes['teamid'] == 2:
                players[pl].defspawn = mapdata.def2
            elif players[pl].attributes['teamid'] == 3:
                players[pl].defspawn = mapdata.def3
            SpawnPlayer(int(pl))

def CreatePlayerPopups():
    '''
Loop through players and create popups for them
Also create spectator popup
    '''
    for pl in players:
        if int(players[pl].attributes['isbot']) == 0:
            if players[pl].attributes['teamid'] > 1:
                CreatePlayerPopup(int(pl))

    CreateSpecPopup()

def UpdatePlayerPopups():
    '''
Loop through players and update popups for them
Also update spectator popup
    '''
    specs = set()
    for pl in players:
        if int(players[pl].attributes['isbot']) == 0:
            team = players[pl].attributes['teamid']
            if team > 1:
                CreatePlayerPopup(int(pl)).send(int(pl))
            elif team == 1:
                specs.add(pl)
    DeleteSpecPopup()
    CreateSpecPopup().send(specs)

def CreatePlayerPopup(userid):
    '''
Create a popup for a player
    '''
    if mapdata:
        team = players[userid].attributes['teamid']
        defs = players[userid].defspawn
        lang = players[userid].get('lang')
        
        p = popuplib.create(macros['playerpopup']%userid)
        p.editlang = lang
        p.addline(textdata('status caption', None, lang))
        p.addline("------------------------")
        for spawn in mapdata.data:
            p.addline(FormSpawnPopupLine(spawn, team, defs))
        p.addline("------------------------")
        p.addline(textdata('status help', None, lang))
        
        p.displaymode = 'sticky'
        p.select(9, SelectSpawnLocationMenu)
        p.submenu(10, macros['otherpopup']%'rules')
        
        popups[macros['playerpopup']%userid] = p
        return p


def CreateSpecPopup():
    '''
Create a popup for spectators
    '''
    if mapdata:
        p = popuplib.create(macros['otherpopup']%'spec')
        
        p.addline(textdata['status caption'])
        p.addline("------------------------")
        for spawn in mapdata.data:
            p.addline(FormSpawnPopupLine(spawn, 1, 0))
        p.addline("------------------------")
        p.addline(textdata['status helpshort'])
        p.displaymode = 'sticky'
        p.submenu(10, macros['otherpopup']%'rules')
        
        popups[macros['otherpopup']%'spec'] = p
        return p


def CreateRulesPopup(rules):
    '''
Create mod rules popup in all languages
    '''
    p = popuplib.create(macros['otherpopup']%'rules')
    if rules['winning'] == 0:
        w = 'rules capture main'
    elif rules['winning'] == 1:
        w = 'rules annihilate'
    elif rules['winning'] == 2:
        w = 'rules capture all'
    
    p.addline(textdata['help caption'])
    p.addline(textdata['rules'])
    p.addline(popuplib.langstring(" - ",textdata[w]))
    
    p.addline(textdata['symbols'])
    p.addline(popuplib.langstring(" > ",textdata['symbol >']))
    p.addline(popuplib.langstring(" ! ",textdata['symbol !']))
    p.addline(popuplib.langstring(" * ",textdata['symbol *']))
    p.addline(textdata['orange lines'])
    
    p.menuselect = ExitRules
    popups[macros['otherpopup']%'rules'] = p
    return p

def ExitRules(userid, choice, popupid):
    '''
Menuselect popup for rules
    '''
    pname = macros['playerpopup']%userid
    popups[pname].send(userid)
    
def StartTracking():
    '''
Start tracking the spawn areas
    '''
    global istracking
    if (not istracking) and mapdata:
        try:
            ut = settings['updatetime'] / len(mapdata.data)
            x = 0.0
            for sp in mapdata.data:
                x += ut
                gamethread.delayed(x, TrackSpawn, mapdata[sp])
        except ZeroDivisionError:
            # empty map, do not track just yet
            pass
        istracking = True

def StopTracking():
    '''
Stop tracking the spawn areas
    '''
    global istracking
    istracking = False

def DeleteRulesPopup():
    '''
Delete mod rules popup
    '''
    pn = macros['otherpopup']%'rules'
    if pn in popups:
        popups[pn].delete()
        del popups[pn]

def DeletePlayerPopups():
    '''
Delete all player popups and spectator popup
    '''
    for pl in players:
        if int(players[pl].attributes['isbot']) == 0:
            DeletePlayerPopup(int(pl))
    DeleteSpecPopup()

def DeletePlayerPopup(userid):
    '''
Delete a popup from a player
    '''
    pname = macros['playerpopup']%userid
    if popups.has_key(pname):
        try:
            popups[pname].unsend(userid)
        except playerlib.UseridError:
            pass
        popups[pname].delete()
        del popups[pname]

def DeleteSpecPopup():
    '''
Delete a spectator popup
    '''
    pname = macros['otherpopup']%'spec'
    if popups.has_key(pname):
        popups[pname].unsend(playerlib.getUseridList("#spec"))
        popups[pname].delete()
        del popups[pname]


def SpawnPlayer(userid):
    '''
Respawn a player to his chosen spawn point
    '''
    if istracking:
        if userid in players:
            team = players[userid].attributes['teamid']
            if team > 1:
                es.server.cmd("est_spawn %d 1"%userid)
                defs = players[userid].defspawn
                location = mapdata[defs].getSpawnLocation()
                es.setpos(userid, location[0], location[1], location[2])

def FormSpawnPopupLine(spawn, team, defspawn):
    '''
Create a line for spawnconquer status popup
    '''
    number = mapdata[spawn].team
    enemyteam = team ^ 1
    colour = ""
    if number == team:
        colour = "->"
    dsmark = ""
    if spawn == defspawn:
        dsmark = "*"
    owncan = " "
    if CanBeCaptured(mapdata[spawn], team):
        owncan = ">"
    enemycan = "    "
    if number == enemyteam:
        if enemyteam == 2:
            enemycan = "(T) "
        elif enemyteam == 3:
            enemycan = "(CT)"
    elif CanBeCaptured(mapdata[spawn], enemyteam):
        enemycan = " !  "
    return "%s%d %s%s %s %s"%(colour, number, owncan, spawn, enemycan, dsmark)

def SelectSpawnLocationMenu(userid, choice, parentname):
    '''
This is a select callback function for selecting default spawn location
    '''
    pname = macros['spawnpopup']%userid
    team = players[userid].attributes['teamid']
    defs = players[userid].defspawn
    lang = players[userid].get("lang")
    
    p = popuplib.easymenu(pname, '_popup_choice', SelectSpawnLocation)
    p.editlang = lang
    p.titleformat = "%s  (%%p/%%t)"%textdata('select spawn', None, lang)
    
    for spawn in mapdata.data:
        if mapdata[spawn].team == team:
            p.addoption(spawn, FormSpawnPopupLine(spawn, team, defs))

    p.submenu(parentname)
    p.send(userid)

def SelectSpawnLocation(userid, choice, parentname):
    '''
This is a menuselectfb callback function for selecting default spawn location
    '''
    popuplib.delete(parentname)
    
    team = players[userid].attributes['teamid']
    spawn = mapdata[choice]
    if spawn.team == team:
        players[userid].defspawn = choice
        CreatePlayerPopup(userid).send(userid)
    else:
        SelectSpawnLocationMenu(userid, 9, parentname)

def CanBeCaptured(spawn, team):
    '''
This function checks if the spawn location can be captured by the team
    '''
    if team < 2:
        return False
    if spawn.team == team:
        return False
    if rules['spawns'] == 0:
        if spawn is mapdata[mapdata.def2]:
            return True
        if spawn is mapdata[mapdata.def3]:
            return True
        return False
    if rules['spawns'] == 1:
        return True
    if rules['spawns'] == 2:
        if team == 2:
            spawnlist = list(spawn.ctlinks)
        if team == 3:
            spawnlist = list(spawn.tlinks)
    elif rules['spawns'] == 3:
        spawnlist = list(spawn.ctlinks)
        spawnlist.extend(spawn.tlinks)
    for sp in spawnlist:
        if mapdata[sp].team == team:
            return True
    return False


def TrackSpawn(spawn):
    '''
Main loop function.
Checks the status of all spawn areas, checks if they are to be captured,
changes their owner if necessary, checks for winning conditions,
draws the spawn areas and the links.
    '''
    if istracking:
        if not inAdminMode:
            spawnteams = {2: 0, 3: 0}
            cteam = spawn.team
            if cteam == 2:
                ccolor = "#red"
            elif cteam == 3:
                ccolor = "#blue"
            else:
                ccolor = "#white"
            inside = set()
            nplayers = playerlib.getPlayerList("#alive")
            for pl in nplayers:
                if spawn.isin(vector(pl.get("location"))+(0,0,32)):
                    inside.add(pl)
                    teamid = pl.get("teamid")
                    spawnteams[teamid] += 1
                    es.toptext(int(pl), 1.0, ccolor, spawn.name)
                    if cteam == teamid:
                        pl.defspawn = spawn.name
            if cteam != 2 and spawnteams[2] > 0 and spawnteams[3] == 0:
                for pl in inside:
                    players[int(pl)].defspawn = spawn.name
                spawn.capture(2)
            elif cteam != 3 and spawnteams[3] > 0 and spawnteams[2] == 0:
                for pl in inside:
                    players[int(pl)].defspawn = spawn.name
                spawn.capture(3)
        elif spawn.name not in mapdata.data:
            # stop tracking self
            return False
        spawn.draw()
        spawn.drawlinks()
        
        gamethread.delayed(settings['updatetime'], TrackSpawn, spawn)

def TriggerWin(team):
    '''
Make a team win and play sounds
    '''
    for pl in players:
        if players[pl].attributes['teamid'] != team:
            es.sexec(int(pl), "kill")
            es.playsound(int(pl), sounds['defeat'], 1.0)
        else:
            es.playsound(int(pl), sounds['victory'], 1.0)
    
def DeleteMapObjectives():
    '''
Delete usual Counter-Strike map objectives for unlimited round time
    '''
    es.dbgmsg(0, "SpawnConquer is deleting map objectives...")
    es.server.cmd("ent_remove_all weapon_c4")
    es.server.cmd("ent_remove_all planted_c4")
    es.server.cmd("ent_remove_all func_bomb_target")
    es.server.cmd("ent_remove_all hostage_entity")
    es.dbgmsg(0, "... done deleting map objectives.")

def DeleteRagdolls(userid):
    '''
Delete all ragdoll entities
    '''
    if int(settings['ragdolls']):
        es.fire(userid, "cs_ragdoll", "kill")

def DeleteWeapons():
    '''
Delete all idle weapon entities
    '''
    if int(settings['weapons']):
        es.server.cmd('est_removeidle weapon_')

def DeleteProps():
    '''
Delete all props from map
    '''
    es.dbgmsg(0, "SpawnConquer is deleting map props...")
    es.server.cmd('ent_remove_all prop_physics')
    es.server.cmd('ent_remove_all prop_physics_multiplayer')
    es.server.cmd('ent_remove_all prop_dynamic')
    es.server.cmd('ent_remove_all prop_ragdoll')
    es.dbgmsg(0, "... done deleting map props.")

def CheckFakeClients():
    '''
Check the status of fake clients, add or remove them if necessary
    '''
    if mapdata:
        if rules['winning'] == 1:
            # No fake clients
            DeleteFakeClients()
        else:
            # Have to be fake clients
            if int(TeamSecurity[2]) == 0:
                # Create them
                es.createbot('_scq_teamsec2', "-T Security-")
                es.createbot('_scq_teamsec3', "-CT Security-")
                es.delayed(0.1,'es_changeteam server_var(_scq_teamsec2) 2')
                es.delayed(0.1,'es_changeteam server_var(_scq_teamsec3) 3')
            # Make sure the fake clients appear as alive
            es.delayed(1,'es_setplayerprop server_var(_scq_teamsec2) CBasePlayer.m_lifeState 256')
            es.delayed(1,'es_setplayerprop server_var(_scq_teamsec3) CBasePlayer.m_lifeState 256')
            # Teleport them somewhere far far away...
            tloc = vector(mapdata[mapdata.def2].origin)
            ctloc = vector(mapdata[mapdata.def3].origin)
            es.delayed(1, 'es_setpos server_var(_scq_teamsec2) %f %f %f'%(tloc[0], tloc[1], -4000))
            es.delayed(1, 'es_setpos server_var(_scq_teamsec3) %f %f %f'%(ctloc[0], ctloc[1], -4000))

def DeleteFakeClients():
    '''
Delete fake clients from the server
    '''
    # Delete existing fake clients
    es.server.cmd("kickid %d"%TeamSecurity[2])
    es.server.cmd("kickid %d"%TeamSecurity[3])
    TeamSecurity[2].set(0)

def AddDownloadables(stuff):
    '''
Add a list of items to be downloaded by clients
    '''
    for item in stuff:
        es.stringtable('downloadables', str(item))

def FindNearestSpawn(loc):
    '''
Finds the spawn area nearest to specified location
    '''
    nearestspawn = None
    spawndistance = 0
    for sp in mapdata.data:
        spawn = mapdata[sp]
        d = vecmath.distance(loc, spawn.origin)
        if not nearestspawn:
            nearestspawn = spawn
            spawndistance = d
        elif d < spawndistance:
            nearestspawn = spawn
            spawndistance = d
    return nearestspawn
    

# ADMIN STUFF FROM HERE

def bullet_impact(event_var):
    '''
This function is called when a bullet hits a wall or object,
only when noisy is enabled
    '''
    global adminEditingCoord, bullet_impact_shield
    if not bullet_impact_shield:
        if adminLinkFrom:
            loc = vector((float(event_var['x']), float(event_var['y']), float(event_var['z'])))
            AdminAddLink2(int(event_var['userid']), loc)
        elif adminAddingSpawn:
            if not adminEditingCoord:
                adminEditingCoord = vector((float(event_var['x']), float(event_var['y']), float(event_var['z'])))
                es.tell(int(event_var['userid']), '1')
            else:
                loc = vector((float(event_var['x']), float(event_var['y']), float(event_var['z'])))
                spwns = len(mapdata.data)
                spawnname = 'spawn %d'%(spwns+1)
                if adminChosenType == 'box':
                    data = 'box;%s;%s'%(str(adminEditingCoord), str(loc))
                elif adminChosenType == 'circle':
                    height = loc[2]-adminEditingCoord[2]
                    if height < 16:
                        height = 65
                    loc[2] = adminEditingCoord[2]
                    radius = vecmath.distance(loc, adminEditingCoord)
                    data = 'circle;%s;%f;%f'%(str(adminEditingCoord),radius,height)
                spawn = SpawnArea(spawnname, data)
                mapdata.data[spawnname] = spawn
                gamethread.delayed(settings['updatetime'], TrackSpawn, spawn)
                es.cexec(int(event_var['userid']), 'slot10')
        elif adminEditingCoord:
            if adminEditingSpawn.type == 'circle' or adminEditingCoord is not adminEditingSpawn.origin:
                adminEditingCoord[0] = float(event_var['x'])
                adminEditingCoord[1] = float(event_var['y'])
                adminEditingCoord[2] = float(event_var['z'])
                adminEditingSpawn.updateOrigin()
            else:
                # edit box origin
                loc = vector((float(event_var['x']), float(event_var['y']), float(event_var['z'])))
                subvector = loc - adminEditingSpawn.origin
                adminEditingSpawn.coord1 += subvector
                adminEditingSpawn.coord2 += subvector
                adminEditingSpawn.origin = loc
                adminEditingCoord = adminEditingSpawn.origin
            p = popups[macros['adminpopup']%'adjust_coord']
            p.modline(2, str(adminEditingCoord))
            p.update(int(event_var['userid']))
            
        elif adminEditingValue == 'radius':
            o = adminEditingSpawn.origin
            loc = vector((float(event_var['x']), float(event_var['y']), o[2]))
            r = vecmath.distance(loc, o)
            adminEditingSpawn.radius = r
            p = popups[macros['adminpopup']%'adjust_value']
            p.modline(2, str(r))
            p.update(int(event_var['userid']))
        elif adminEditingValue == 'height':
            o = adminEditingSpawn.origin
            h = abs(float(event_var['z']) - o[2])
            adminEditingSpawn.height = h
            p = popups[macros['adminpopup']%'adjust_value']
            p.modline(2, str(h))
            p.update(int(event_var['userid']))
        bullet_impact_shield = True
        gamethread.queue(DisableBulletImpactShield)

def DisableBulletImpactShield():
    '''
Prevents double bullet_impact firing
    '''
    global bullet_impact_shield
    bullet_impact_shield = False

def chat_monitor(userid, text, teamonly):
    '''
This function is called each time a player says something in text chat
This will only be used for admin functions now
    '''
    text = text.strip('"')
    global inAdminMode, adminEditingText
    if text == '!scqadmin':
        auth = services.use("auth")
        if auth.isUseridAuthorized(userid, "edit_spawnconquer"):
            inAdminMode = True
            startAdminMode(userid)
        else:
            lang = playerlib.getPlayer(userid).get('lang')
            es.tell(userid, textdata('not authorized', {'command': text}, lang))
        return (0, None, None)
    if not inAdminMode:
        # Show the chat message
        return (userid, text, teamonly)
    lang = playerlib.getPlayer(userid).get('lang')
    if adminEditingText == userid:
        oldname = adminEditingSpawn.name
        result = adminEditingSpawn.rename(text)
        es.tell(userid, textdata('admin action rename%s'%str(result), {
            'old': oldname,
            'new': text
            }, lang))
        adminEditingText = None
        return (0, None, None)
    return (userid, text, teamonly)

def startAdminMode(userid):
    '''
Start admin mode for user
    '''
    global mapdata
    AdminPopupsFor(userid)
    popups[macros['adminpopup']%'main'].send(userid)
    if not mapdata:
        mapdata = Mapdata(
            {'author': 'unknown', 'version': 'unknown', 's_version': info.version},
            "",
            ""
            )
    StartTracking()
    es.server.cmd('bot_kick')
    es.soon('est_stripplayer %d 1'%userid)
    es.soon('es_give %d weapon_usp'%userid)
    es.soon('es_setplayerprop %d CBasePlayer.localdata.m_iAmmo.008 1000'%userid)
    es.soon('est_god %d 1'%userid)

def AdminPopupsFor(userid):
    '''
Generate the admin popups for the user
    '''
    user = playerlib.getPlayer(userid)
    lang = user.get('lang')

    # Main menu
    
    pname = macros['adminpopup']%'main'
    p = popuplib.create(pname)
    popups[pname] = p
    p.addline(textdata('admin main', None, lang))
    p.addline("->1. %s"%textdata('admin main add', None, lang))
    p.addline("->2. %s"%textdata('admin main edit', None, lang))
    p.addline("->3. %s"%textdata('admin main delete', None, lang))
    p.addline(" ")
    p.addline("->4. %s [%s]"%
                  (
                  textdata('admin main inverse', None, lang),
                  textdata('choice %s'%str(adminInverseLinks), None, lang)
                  )
              )
    p.addline("->5. %s"%textdata('admin main addt', None, lang))
    p.addline("->6. %s"%textdata('admin main addct', None, lang))
    p.addline("->7. %s"%textdata('admin main adduni', None, lang))
    p.addline(" ")
    p.addline("->8. %s"%textdata('admin main maket', None, lang))
    p.addline("->9. %s"%textdata('admin main makect', None, lang))
    p.addline(" ")
    p.addline("0. %s"%textdata('admin main file', None, lang))

    
    p.submenu(3,  pname)
    p.submenu(4,  pname)
    p.submenu(5,  macros['adminpopup']%'addlink')
    p.submenu(6,  macros['adminpopup']%'addlink')
    p.submenu(7,  macros['adminpopup']%'addlink')
    p.submenu(8,  pname)
    p.submenu(9,  pname)
    p.submenu(10, macros['adminpopup']%'fileopts')

    p.select(1, AdminCreateAddSpawnPopup)
    p.select(2, AdminEditSpawn)
    p.select(3, AdminDeleteSpawn)
    p.select(4, AdminSwapAutomaticInverseLinks)
    p.select(5, AdminAddLink1)
    p.select(6, AdminAddLink1)
    p.select(7, AdminAddLink1)
    p.select(8, AdminMakeDefT)
    p.select(9, AdminMakeDefCT)
    
    
    # Edit spawn (box)
    
    pname = macros['adminpopup']%'editspawn_box'
    p = popuplib.create(pname)
    popups[pname] = p
    p.addline("%s (%s)"%
              (
                  textdata('admin edit', None, lang),
                  textdata('spawntype box', None, lang)
               )
              )
    p.addline("->1. %s"%textdata('admin edit corner1', None, lang))
    p.addline("->2. %s"%textdata('admin edit corner2', None, lang))
    p.addline("->3. %s"%textdata('admin edit movebox', None, lang))
    p.addline("->6. %s"%textdata('admin edit rename', None, lang))
    p.addline("->7. %s"%textdata('admin edit clear', None, lang))
    p.addline("->8. %s"%textdata('admin edit clearall', None, lang))
    p.addline("->9. %s (->%s)"%
              (
                  textdata('admin edit changetype', None, lang),
                  textdata('spawntype circle', None, lang)
              )
             )
    p.addline(" ")
    p.addline("0. %s"%textdata('admin edit cancel', None, lang))

    p.select(1, AdminEditBoxCorner)
    p.select(2, AdminEditBoxCorner)
    p.select(3, AdminEditBoxCorner)
    p.select(6, AdminRenameSpawn)
    p.submenu(6, pname)
    p.select(7, AdminClearLinks)
    p.submenu(7, pname)
    p.select(8, AdminClearAllLinks)
    p.submenu(8, pname)
    p.select(9, AdminChangeType)
    p.select(10, AdminStopSpawnEdit)
    p.submenu(10, macros['adminpopup']%'main')

    # Edit spawn (circle)
    
    pname = macros['adminpopup']%'editspawn_circle'
    p = popuplib.create(pname)
    popups[pname] = p
    p.addline("%s (%s)"%
              (
                  textdata('admin edit', None, lang),
                  textdata('spawntype circle', None, lang)
               )
              )
    p.addline("->1. %s"%textdata('admin edit origin', None, lang))
    p.addline("->2. %s"%textdata('admin edit radius', None, lang))
    p.addline("->3. %s"%textdata('admin edit height', None, lang))
    p.addline("->6. %s"%textdata('admin edit rename', None, lang))
    p.addline("->7. %s"%textdata('admin edit clear', None, lang))
    p.addline("->8. %s"%textdata('admin edit clearall', None, lang))
    p.addline("->9. %s (->%s)"%
              (
                  textdata('admin edit changetype', None, lang),
                  textdata('spawntype box', None, lang)
              )
             )
    p.addline(" ")
    p.addline("0. %s"%textdata('admin edit cancel', None, lang))
    
    p.select(1, AdminEditOrigin)
    p.select(2, AdminEditRadius)
    p.select(3, AdminEditHeight)
    p.select(6, AdminRenameSpawn)
    p.submenu(6, pname)
    p.select(7, AdminClearLinks)
    p.submenu(7, pname)
    p.select(8, AdminClearAllLinks)
    p.submenu(8, pname)
    p.select(9, AdminChangeType)
    p.select(10, AdminStopSpawnEdit)
    p.submenu(10, macros['adminpopup']%'main')

    # File options
    
    pname = macros['adminpopup']%'fileopts'
    p = popuplib.create(pname)
    popups[pname] = p
    p.addline(textdata('admin file', None, lang))
    p.addline("->1. %s"%textdata('admin file clear', None, lang))
    p.addline("->2. %s"%textdata('admin file reload', None, lang))
    p.addline("->3. %s"%textdata('admin file save', None, lang))
    p.addline("->4. %s"%textdata('admin file finish', None, lang))
    p.addline(" ")
    p.addline("0. %s"%textdata('admin file cancel', None, lang))
    
    p.select(1, AdminFileClear)
    p.select(2, AdminFileReload)
    p.select(3, AdminFileSave)
    p.select(4, AdminFileFinish)
    
    p.submenu(1,  macros['adminpopup']%'main')
    p.submenu(2,  macros['adminpopup']%'main')
    p.submenu(3,  pname)
    p.submenu(10, macros['adminpopup']%'main')

    # Create link

    pname = macros['adminpopup']%'addlink'
    p = popuplib.create(pname)
    popups[pname] = p
    p.addline(textdata('admin action linkhelp', None, lang))
    p.menuselect = AdminStopLinking
    
    # Adjust value
    
    pname = macros['adminpopup']%'adjust_value'
    p = popuplib.create(pname)
    popups[pname] = p
    p.addline(textdata('admin adjust', None, lang))
    p.addline("<value>")
    p.addline(textdata('admin adjust instructions', None, lang))
    p.addline("->1. %s + %d"%
              (
                  textdata('admin adjust inc', None, lang),
                  adminAdjustStep
              )
             )
    p.addline("->2. %s - %d"%
              (
                  textdata('admin adjust dec', None, lang),
                  adminAdjustStep
              )
             )
    p.addline("->3. %s"%textdata('admin adjust incstep', None, lang))
    p.addline("->4. %s"%textdata('admin adjust decstep', None, lang))
    p.addline(" ")
    p.addline("0. %s"%textdata('admin adjust cancel', None, lang))
    
    p.select(1, AdminIncreaseValue)
    p.select(2, AdminDecreaseValue)
    p.select(3, AdminIncreaseStep)
    p.select(4, AdminDecreaseStep)
    p.select(10, AdminStopValue)
    
    p.submenu(1,  pname)
    p.submenu(2,  pname)
    p.submenu(3,  pname)
    p.submenu(4,  pname)
    #p.submenu(10, whateverthepreviousmenuis)
    
    # Adjust coordinate
    
    pname = macros['adminpopup']%'adjust_coord'
    p = popuplib.create(pname)
    popups[pname] = p
    p.addline(textdata('admin adjust', None, lang))
    p.addline("<value>")
    p.addline(textdata('admin adjust instructions', None, lang))
    p.addline("->1. x + %d"%adminAdjustStep)
    p.addline("->2. x - %d"%adminAdjustStep)
    p.addline("->3. y + %d"%adminAdjustStep)
    p.addline("->4. y - %d"%adminAdjustStep)
    p.addline("->5. z + %d"%adminAdjustStep)
    p.addline("->6. z - %d"%adminAdjustStep)
    p.addline("->7. %s"%textdata('admin adjust incstep', None, lang))
    p.addline("->8. %s"%textdata('admin adjust decstep', None, lang))
    p.addline("->9. %s"%textdata('admin adjust reset', None, lang))
    p.addline(" ")
    p.addline("0. %s"%textdata('admin adjust cancel', None, lang))
    
    p.select(1, AdminIncreaseCoord)
    p.select(2, AdminDecreaseCoord)
    p.select(3, AdminIncreaseCoord)
    p.select(4, AdminDecreaseCoord)
    p.select(5, AdminIncreaseCoord)
    p.select(6, AdminDecreaseCoord)
    p.select(7, AdminIncreaseStep)
    p.select(8, AdminDecreaseStep)
    p.select(9, AdminResetCoord)
    p.select(10, AdminStopCoord)
    
    p.submenu(1,  pname)
    p.submenu(2,  pname)
    p.submenu(3,  pname)
    p.submenu(4,  pname)
    p.submenu(5,  pname)
    p.submenu(6,  pname)
    p.submenu(7,  pname)
    p.submenu(8,  pname)
    p.submenu(9,  pname)
    #p.submenu(10, whateverthepreviousmenuis)

def AdminCreateAddSpawnPopup(userid, choice, popupid):
    '''
Add spawn
    '''
    global adminAddingSpawn
    
    if not adminAddingSpawn:
        es.doblock('corelib/noisy_on')
        adminAddingSpawn = True
    
    lang = playerlib.getPlayer(userid).get('lang')
    
    pname = macros['adminpopup']%'addspawn'
    p = popuplib.create(pname)
    popups[pname] = p
    p.addline(textdata('admin add', None, lang))
    p.addline(textdata('admin add choose', None, lang))
    if adminChosenType == 'box':
        p.addline("->1. %s (X)"%textdata('spawntype box', None, lang))
    else:
        p.addline("->1. %s ( )"%textdata('spawntype box', None, lang))
    if adminChosenType == 'circle':
        p.addline("->2. %s (X)"%textdata('spawntype circle', None, lang))
    else:
        p.addline("->2. %s ( )"%textdata('spawntype circle', None, lang))
    p.addline(textdata('admin add instructions', None, lang))
    p.addline("0. %s"%textdata('admin add cancel', None, lang))

    p.menuselect = AdminChooseType
    p.send(userid)


def AdminDeleteSpawn(userid, choice, popupid):
    '''
Admin function: delete the spawn area closest to the admin
    '''
    playerloc = vector(es.getplayerlocation(userid))
    lang = playerlib.getPlayer(userid).get('lang')
    spawn = FindNearestSpawn(playerloc)
    if spawn:
        spawn.clearLinks()
        spawn.clearToLinks()
        name = str(spawn.name)
        del mapdata.data[name]
        del spawn
        es.tell(userid, textdata('admin action deleted', {'spawn': name}, lang))

def AdminSwapAutomaticInverseLinks(userid, choice, popupid):
    '''
Change the state of creating automatically back links for other team
    '''
    global adminInverseLinks
    lang = playerlib.getPlayer(userid).get('lang')
    adminInverseLinks = not adminInverseLinks
    p = popups[macros['adminpopup']%'main']
    p.modline(6, "->4. %s [%s]"%
                  (
                  textdata('admin main inverse', None, lang),
                  textdata('choice %s'%str(adminInverseLinks), None, lang)
                  )
              )

def AdminAddLink1(userid, choice, popupid):
    '''
Start linking from current spawn and end to second one
    '''
    global adminLinkFrom, adminLinkTeam
    playerloc = vector(es.getplayerlocation(userid))
    lang = playerlib.getPlayer(userid).get('lang')
    spawn = FindNearestSpawn(playerloc)
    if spawn:
        adminLinkFrom = spawn
        adminLinkTeam = int(choice) - 3
        name = spawn.name
        es.doblock('corelib/noisy_on')
        es.tell(userid, textdata('admin action startlink', {'spawn': name}, lang))

def AdminStopLinking(userid, choice, popupid):
    '''
Stop linking process and send the main menu again
    '''
    global adminLinkFrom
    pname = macros['adminpopup']%'main'
    popups[pname].send(userid)
    es.doblock('corelib/noisy_off')
    adminLinkFrom = None

def AdminAddLink2(userid, targetloc):
    '''
Finish linking
    '''
    spawn = FindNearestSpawn(targetloc)
    lang = playerlib.getPlayer(userid).get('lang')
    if spawn is adminLinkFrom:
        es.tell(userid, textdata('admin action selflink', None, lang))
    else:
        # we have a real link
        if adminLinkTeam == 2 or adminLinkTeam == 4:
            adminLinkFrom.tlinks.add(spawn.name)
            es.tell(userid, textdata('admin action linkedt', {
                'source': adminLinkFrom.name,
                'target': spawn.name
                }, lang))
            if adminInverseLinks:
                spawn.ctlinks.add(adminLinkFrom.name)
                es.tell(userid, textdata('admin action linkedct', {
                    'source': spawn.name,
                    'target': adminLinkFrom.name
                    }, lang))
        if adminLinkTeam == 3 or adminLinkTeam == 4:
            adminLinkFrom.ctlinks.add(spawn.name)
            es.tell(userid, textdata('admin action linkedct', {
                'source': adminLinkFrom.name,
                'target': spawn.name
                }, lang))
            if adminInverseLinks:
                spawn.tlinks.add(adminLinkFrom.name)
                es.tell(userid, textdata('admin action linkedt', {
                    'source': spawn.name,
                    'target': adminLinkFrom.name
                    }, lang))

    es.cexec(userid, 'slot10')

def AdminMakeDefT(userid, choice, popupid):
    '''
Make the nearest spawn area the Terrorist map spawn
    '''
    playerloc = vector(es.getplayerlocation(userid))
    lang = playerlib.getPlayer(userid).get('lang')
    spawn = FindNearestSpawn(playerloc)
    if spawn:
        es.tell(userid, textdata('admin action makedeft', {'spawn': spawn.name}, lang))
        if mapdata.def2:
            mapdata[mapdata.def2].team = 0
        mapdata.def2 = spawn.name
        spawn.team = 2


def AdminMakeDefCT(userid, choice, popupid):
    '''
Make the nearest spawn area the Counter-Terrorist map spawn
    '''
    playerloc = vector(es.getplayerlocation(userid))
    lang = playerlib.getPlayer(userid).get('lang')
    spawn = FindNearestSpawn(playerloc)
    if spawn:
        es.tell(userid, textdata('admin action makedefct', {'spawn': spawn.name}, lang))
        if mapdata.def3:
            mapdata[mapdata.def3].team = 0
        mapdata.def3 = spawn.name
        spawn.team = 3


def AdminChooseType(userid, choice, popupid):
    '''
Admin chose a new spawn type from the add spawn popup
    '''
    global adminChosenType, adminAddingSpawn, adminEditingCoord
    if choice == 1:
        adminChosenType = 'box'
    elif choice == 2:
        adminChosenType = 'circle'

    if choice < 10:
        AdminCreateAddSpawnPopup(userid, 1, None)
    else:
        adminAddingSpawn = False
        adminEditingCoord = None
        es.doblock('corelib/noisy_off')
        popups[macros['adminpopup']%'main'].send(userid)

def AdminEditSpawn(userid, choice, popupid):
    '''
Enter edit spawn menu
    '''
    global adminEditingSpawn
    playerloc = vector(es.getplayerlocation(userid))
    lang = playerlib.getPlayer(userid).get('lang')
    spawn = FindNearestSpawn(playerloc)
    if spawn:
        adminEditingSpawn = spawn
        spawn.team = 1
        t = spawn.type
        pn = macros['adminpopup']%'editspawn_%s'%t
        popups[pn].send(userid)
    else:
        pn = macros['adminpopup']%'main'
        popups[pn].send(userid)

        
def AdminStopSpawnEdit(userid, choice, popupid):
    '''
Exit edit spawn menu
    '''
    global adminEditingSpawn, adminEditingCoord
    if adminEditingSpawn:
        if mapdata.def2 == adminEditingSpawn.name:
            adminEditingSpawn.team = 2
        elif mapdata.def3 == adminEditingSpawn.name:
            adminEditingSpawn.team = 3
        else:
            adminEditingSpawn.team = 0
        adminEditingSpawn = None
        adminEditingCoord = None
        
def AdminEditBoxCorner(userid, choice, popupid):
    '''
Start moving a box corner
    '''
    global adminEditingCoord, adminCoordBackup
    if choice == 1:
        adminEditingCoord = adminEditingSpawn.coord1
    elif choice == 2:
        adminEditingCoord = adminEditingSpawn.coord2
    elif choice == 3:
        adminEditingCoord = adminEditingSpawn.origin
    else:
        raise Exception
    adminCoordBackup = vector(adminEditingCoord)
    
    pname = macros['adminpopup']%'adjust_coord'
    p = popups[pname]
    p.modline(2, str(adminEditingCoord))
    p.submenu(10, popupid)
    p.send(userid)
    es.doblock('corelib/noisy_on')

def AdminStopCoord(userid, choice, popupid):
    '''
Stop editing a box corner
    '''
    es.doblock('corelib/noisy_off')
    
def AdminRenameSpawn(userid, choice, popupid):
    '''
Rename spawn area
    '''
    global adminEditingText
    lang = playerlib.getPlayer(userid).get('lang')
    es.tell(userid, textdata('admin action rename', None, lang))
    adminEditingText = userid
    
def AdminClearLinks(userid, choice, popupid):
    '''
Clear all links coming from the current spawn area
    '''
    adminEditingSpawn.clearLinks()

def AdminClearAllLinks(userid, choice, popupid):
    '''
Clear all links from the current spawn and pointing to current spawn
    '''
    adminEditingSpawn.clearToLinks()
    
def AdminChangeType(userid, choice, popupid):
    '''
Change a spawn area's type between box and circle
    '''
    if adminEditingSpawn.type == 'box':
        # change to circle
        adminEditingSpawn.type = 'circle'
        adminEditingSpawn.height = abs(adminEditingSpawn.coord1[2]-adminEditingSpawn.coord2[2])
        adminEditingSpawn.radius = vecmath.distance(adminEditingSpawn.coord1, adminEditingSpawn.coord2)/2
        pname = macros['adminpopup']%'editspawn_circle'
        popups[pname].send(userid)
    elif adminEditingSpawn.type == 'circle':
        # change to box
        adminEditingSpawn.type = 'box'
        v1 = vector(adminEditingSpawn.origin)
        v2 = vector(v1)
        v1[0] = v1[0] - pow(adminEditingSpawn.radius, .5)
        v1[1] = v1[1] - pow(adminEditingSpawn.radius, .5)
        v2[0] = v2[0] + pow(adminEditingSpawn.radius, .5)
        v2[1] = v2[1] + pow(adminEditingSpawn.radius, .5)
        v2[2] = v2[2] + adminEditingSpawn.height
        adminEditingSpawn.coord1 = v1
        adminEditingSpawn.coord2 = v2

def AdminEditOrigin(userid, choice, popupid):
    '''
Edit the origin of a circle
    '''
    AdminEditBoxCorner(userid, 3, popupid)


def AdminEditRadius(userid, choice, popupid):
    '''
Start altering the radius of circle spawn
    '''
    global adminEditingValue
    adminEditingValue = 'radius'
    pname = macros['adminpopup']%'adjust_value'
    p = popups[pname]
    p.modline(2, str(adminEditingSpawn.radius))
    p.submenu(10, popupid)
    p.send(userid)
    es.doblock('corelib/noisy_on')

def AdminEditHeight(userid, choice, popupid):
    '''
Start altering the height of circle spawn
    '''
    global adminEditingValue
    adminEditingValue = 'height'
    pname = macros['adminpopup']%'adjust_value'
    p = popups[pname]
    p.modline(2, str(adminEditingSpawn.height))
    p.submenu(10, popupid)
    p.send(userid)
    es.doblock('corelib/noisy_on')

def AdminStopValue(userid, choice, popupid):
    '''
Stop altering radius/height
    '''
    global adminEditingValue
    adminEditingValue = None


def AdminFileClear(userid, choice, popupid):
    '''
Clear all map data!
    '''
    lang = playerlib.getPlayer(userid).get('lang')
    mapdata.clear()
    es.tell(userid, textdata('admin action map_clear', None, lang))
    
def AdminFileReload(userid, choice, popupid):
    '''
Reload map data!
    '''
    global mapdata
    lang = playerlib.getPlayer(userid).get('lang')
    mapdata = LoadMapdata(str(currentmap))
    es.tell(userid, textdata('admin action map_load', None, lang))
    
def AdminFileSave(userid, choice, popupid):
    '''
Save the map data
    '''
    user = playerlib.getPlayer(userid)
    lang = user.get('lang')
    mapdata.author = str(user.get('name'))
    mapdata.version = time.strftime('%Y-%m-%dT%H:%M')
    SaveMapdata(mapdata, str(currentmap))
    es.tell(userid, textdata('admin action map_save', None, lang))
    
def AdminFileFinish(userid, choice, popupid):
    '''
Restart the map
    '''
    es.server.cmd('changelevel %s'%str(currentmap))
    

def AdminIncreaseValue(userid, choice, popupid):
    '''
Increase the value of currently edited thing
    '''
    if adminEditingValue == 'radius':
        adminEditingSpawn.radius += adminAdjustStep
        pname = macros['adminpopup']%'adjust_value'
        p = popups[pname]
        p.modline(2, str(adminEditingSpawn.radius))
        p.update(userid)
    if adminEditingValue == 'height':
        adminEditingSpawn.height += adminAdjustStep
        pname = macros['adminpopup']%'adjust_value'
        p = popups[pname]
        p.modline(2, str(adminEditingSpawn.height))
        p.update(userid)


def AdminDecreaseValue(userid, choice, popupid):
    '''
Decrease the value of currently edited thing
    '''
    if adminEditingValue == 'radius':
        adminEditingSpawn.radius -= adminAdjustStep
        pname = macros['adminpopup']%'adjust_value'
        p = popups[pname]
        p.modline(2, str(adminEditingSpawn.radius))
        p.update(userid)
    if adminEditingValue == 'height':
        adminEditingSpawn.height -= adminAdjustStep
        pname = macros['adminpopup']%'adjust_value'
        p = popups[pname]
        p.modline(2, str(adminEditingSpawn.height))
        p.update(userid)

def AdminIncreaseStep(userid, choice, popupid):
    '''
Increase editing step
    '''
    global adminAdjustStep
    adminAdjustStep += 1
    if adminAdjustStep > 20:
        adminAdjustStep = 1
    AdminStepUpdate(userid)

def AdminDecreaseStep(userid, choice, popupid):
    '''
Decrease editing step
    '''
    global adminAdjustStep
    adminAdjustStep -= 1
    if adminAdjustStep < 1:
        adminAdjustStep = 20
    AdminStepUpdate(userid)


def AdminStepUpdate(userid):
    '''
Update popups with new step values
    '''
    lang = playerlib.getPlayer(userid).get('lang')
    if adminEditingValue:
        pname = macros['adminpopup']%'adjust_value'
        p = popups[pname]
        p.modline(4, "->1. %s + %d"%
                  (
                      textdata('admin adjust inc', None, lang),
                      adminAdjustStep
                  )
                 )
        p.modline(5, "->2. %s - %d"%
                  (
                      textdata('admin adjust dec', None, lang),
                      adminAdjustStep
                  )
                 )
        p.update(userid)
    elif adminEditingCoord:
        pname = macros['adminpopup']%'adjust_coord'
        p = popups[pname]
        p.modline(4, "->1. x + %d"%adminAdjustStep)
        p.modline(5, "->2. x - %d"%adminAdjustStep)
        p.modline(6, "->3. y + %d"%adminAdjustStep)
        p.modline(7, "->4. y - %d"%adminAdjustStep)
        p.modline(8, "->5. z + %d"%adminAdjustStep)
        p.modline(9, "->6. z - %d"%adminAdjustStep)
        p.update(userid)


def AdminIncreaseCoord(userid, choice, popupid):
    '''
Increase the coordinate value
    '''
    ci = int((choice-1)/2)
    adminEditingCoord[ci] += adminAdjustStep
    if adminEditingSpawn.type == 'box' and adminEditingCoord is adminEditingSpawn.origin:
        adminEditingSpawn.coord1[ci] += adminAdjustStep
        adminEditingSpawn.coord2[ci] += adminAdjustStep
    else:
        adminEditingSpawn.updateOrigin()
    p = popups[macros['adminpopup']%'adjust_coord']
    p.modline(2, str(adminEditingCoord))
    p.update(userid)

def AdminDecreaseCoord(userid, choice, popupid):
    '''
Decrease the coordinate value
    '''
    ci = int((choice-1)/2)
    adminEditingCoord[ci] -= adminAdjustStep
    if adminEditingSpawn.type == 'box' and adminEditingCoord is adminEditingSpawn.origin:
        adminEditingSpawn.coord1[ci] -= adminAdjustStep
        adminEditingSpawn.coord2[ci] -= adminAdjustStep
    else:
        adminEditingSpawn.updateOrigin()
    p = popups[macros['adminpopup']%'adjust_coord']
    p.modline(2, str(adminEditingCoord))
    p.update(userid)

def AdminResetCoord(userid, choice, popupid):
    '''
Reset the coordinate value
    '''
    adminEditingCoord[0] = adminCoordBackup[0]
    adminEditingCoord[1] = adminCoordBackup[1]
    adminEditingCoord[2] = adminCoordBackup[2]
    adminEditingSpawn.updateOrigin()
    p = popups[macros['adminpopup']%'adjust_coord']
    p.modline(2, str(adminEditingCoord))
    p.update(userid)

